from typing import Dict, Union

from bolinette import services, db, mapping

_registered_services: Dict[str, 'services.BaseService'] = {}


def register(service: Union['services.BaseService', 'services.SimpleService']):
    _registered_services[service.name] = service


def get(name: str):
    return _registered_services[name]


def init_services():
    for name, service in _registered_services.items():
        if isinstance(service, services.BaseService):
            service.model = db.models.get(name)
            if issubclass(service.model, db.defs.Model):
                mapping.register(name, service.model)



from bolinette import mapping, db


@db.model('file')
class File(db.defs.Model):
    id = db.defs.Column(db.types.Integer, primary_key=True)
    key = db.defs.Column(db.types.String, nullable=False)
    name = db.defs.Column(db.types.String, nullable=False)
    mime = db.defs.Column(db.types.String, nullable=False)

    @classmethod
    def responses(cls):
        yield [
            mapping.Column(cls.key),
            mapping.Column(cls.name),
            mapping.Column(cls.mime)
        ]
        yield 'minimal', [
            mapping.Column(cls.key)
        ]



from typing import Pattern, List, Any, Dict
import re

from aiohttp.web_ws import WebSocketResponse

from bolinette import ws
from bolinette.network import transaction


class Channel:
    def __init__(self, name: str, rule: Pattern[str], function):
        self.name = name
        self.rule = re.compile(f'^{rule}$')
        self.function = function


class Topic:
    def __init__(self, name: str, *, login_required=False, roles=None):
        self.name = name
        self.login_required = login_required
        self.roles = roles or []
        self._channels = []
        self._subscription_func = self._default_subscription_func
        self._subscriptions: Dict[str, List[WebSocketResponse]] = {}
        ws.resources.register_topic(self)

    @property
    def subscriptions(self):
        return self._subscriptions

    @staticmethod
    def _default_subscription_func(topic, response: WebSocketResponse, *, payload: Dict[str, Any], **_):
        channel_name = payload['channel']
        if channel_name not in topic.subscriptions:
            topic.subscriptions[channel_name] = []
        topic.subscriptions[channel_name].append(response)

    async def _receive_message(self, payload: Dict[str, Any], current_user):
        for channel in self._channels:
            if channel.rule.match(payload['channel']):
                await channel.function(self, payload=payload, current_user=current_user)

    async def _receive_subscription(self, response: WebSocketResponse, payload: Dict[str, Any], current_user):
        self._subscription_func(self, response, payload=payload, current_user=current_user)

    async def send_message(self, channels: List[str], message: Any):
        for channel in channels:
            if any(filter(lambda r: r.closed, self._subscriptions[channel])):
                self._subscriptions[channel] = list(filter(lambda r: not r.closed, self._subscriptions[channel]))
            for response in self._subscriptions[channel]:
                await response.send_json({
                    'topic': self.name,
                    'channel': channel,
                    'message': message
                })

    async def process(self, response: WebSocketResponse, payload, current_user):
        if self.login_required and current_user is None:
            return
        action = payload['action']
        with transaction:
            if action == 'send':
                await self._receive_message(payload, current_user)
            elif action == 'subscribe':
                await self._receive_subscription(response, payload, current_user)

    def subscribe(self, function):
        self._subscription_func = function
        return function

    def channel(self, rule: Pattern[str]):
        def decorator(function):
            self._channels.append(Channel(function.__name__, rule, function))
            return function

        return decorator



import string
import random

from bolinette import env
from bolinette.services import BaseService
from bolinette.utils import fs


class FileService(BaseService):
    def __init__(self):
        super().__init__('file')

    async def _generate_key(self):
        key = None
        while key is None or len(await self.get_by('key', key)):
            key = ''.join(random.choices(string.ascii_lowercase + string.digits, k=32))
        return key

    async def write_file(self, content, key):
        path = env.instance_path('uploads')
        if not fs.exists(path):
            fs.mkdir(path)
        with open(fs.join(path, key), 'wb') as f:
            f.write(content)

    async def delete_file(self, key):
        fs.delete(env.instance_path('uploads', key))

    async def delete(self, entity, **_):
        ent = await super().delete(entity)
        await self.delete_file(entity.key)
        return ent

    async def save_file(self, request_file):
        params = {
            'key': await self._generate_key(),
            'name': request_file.filename,
            'mime': request_file.content_type
        }
        await self.write_file(request_file.file.read(), params['key'])
        return await self.create(params)

    async def file_sender(self, key):
        with open(env.instance_path('uploads', key), 'rb') as f:
            chunk = f.read(2 ** 16)
            while chunk:
                yield chunk
                chunk = f.read(2 ** 16)


file_service = FileService()



from bolinette.web.response import response, Cookie
from bolinette.web.resources import resources
from bolinette.web.serializing import deserialize, serialize
from bolinette.web.method import Method
from bolinette.web.route import Route
from bolinette.web.namespace import Namespace



from aiohttp import web as aio_web

from bolinette import env
from bolinette.web import resources
from bolinette.commands import command


@command('run_server')
def run_server():
    aio_web.run_app(resources.app, port=env['PORT'])



from typing import Dict, List

import aiohttp
from aiohttp import web as aio_web

from bolinette import web, ws, env
from bolinette.exceptions import APIError
from bolinette.network import AccessToken
from bolinette.services import user_service
from bolinette.utils import console, logger


class Resources:
    def __init__(self):
        self._topics: Dict[str, ws.Topic] = {}
        self._anon_responses: List[aio_web.WebSocketResponse] = []
        self._responses: Dict[str, aio_web.WebSocketResponse] = {}

    def _register_response(self, identity: str, response: aio_web.WebSocketResponse):
        self._responses[identity] = response

    def _unregister_response(self, identity):
        del self._responses[identity]

    def _register_anon_response(self, response: aio_web.WebSocketResponse):
        self._anon_responses.append(response)

    def _unregister_anon_response(self, response: aio_web.WebSocketResponse):
        self._anon_responses.remove(response)

    def get_response(self, identity):
        return self._responses.get(identity)

    def register_topic(self, topic: 'ws.Topic'):
        self._topics[topic.name] = topic

    def __getitem__(self, key):
        return self._topics.get(key)

    async def handler(self, request):
        response = aio_web.WebSocketResponse()
        await response.prepare(request)
        logger.warning('New WS connection')

        session = None
        identity = AccessToken.Optional.check(request)
        if identity is not None:
            session = await user_service.get_by_username(identity)
        if session is not None:
            self._register_response(session.username, response)
        else:
            self._register_anon_response(response)

        async for msg in response:  # type: aiohttp.WSMessage
            try:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    payload = msg.json()
                    action = payload['action']
                    if action == 'close':
                        await response.close()
                    elif action == 'ping':
                        await response.send_str('pong')
                    else:
                        await self._topics[payload['topic']].process(response, payload, session)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    console.print(f'ws connection closed with exception {response.exception()})')
            except APIError as e:
                for message in e.messages:
                    logger.error(message)

        if session is not None:
            self._unregister_response(session.username)
        else:
            self._unregister_anon_response(response)

        logger.warning('Closed WS connection')
        return response

    def init_app(self):
        env.topics = self
        web.resources.app.add_routes([aio_web.get('/ws', self.handler)])


resources = Resources()



from typing import Union, Dict, List, Tuple

from bolinette import db

MappingPyTyping = List[Union['mapping.Column', 'mapping.Field', 'mapping.List', 'mapping.Definition']]
MappingListPyTyping = Union[MappingPyTyping, Tuple[str, MappingPyTyping]]


class Mixin:
    @staticmethod
    def columns() -> Dict[str, 'db.defs.Column']:
        pass

    @staticmethod
    def relationships(model_cls) -> Dict[str, 'db.defs.Relationship']:
        pass


class Model:
    __orm_model__ = None
    __model_name__ = None

    @staticmethod
    def _get_attribute_of_type(cls, attr_type):
        return dict([(name, attribute) for name, attribute in vars(cls).items() if isinstance(attribute, attr_type)])

    @classmethod
    def payloads(cls) -> MappingListPyTyping:
        pass

    @classmethod
    def responses(cls) -> MappingListPyTyping:
        pass

    @classmethod
    def get_columns(cls) -> Dict[str, 'Column']:
        return cls._get_attribute_of_type(cls, Column)

    @classmethod
    def get_relationships(cls) -> Dict[str, 'Relationship']:
        return cls._get_attribute_of_type(cls, Relationship)

    @classmethod
    def get_properties(cls) -> Dict[str, 'ModelProperty']:
        return cls._get_attribute_of_type(cls, ModelProperty)

    @classmethod
    def query(cls):
        return db.engine.session.query(cls.__orm_model__)

    def __new__(cls, *args, **kwargs):
        return cls.__orm_model__(*args, **kwargs)


class ModelProperty:
    def __init__(self, name, function):
        self.name = name
        self.function = function

    def __repr__(self):
        return f'<ModelProperty {self.name}>'


class Reference:
    def __init__(self, model_name: str, column_name: str):
        self.model_name = model_name
        self.column_name = column_name

    def __repr__(self):
        return f'<Reference -> {self.model_name}.{self.column_name}'


class Column:
    def __init__(self, data_type: 'db.types.DataType', *, reference: Reference = None, primary_key: bool = False,
                 nullable: bool = True, unique: bool = False):
        self.type = data_type
        self.reference = reference
        self.primary_key = primary_key
        self.nullable = nullable
        self.unique = unique
        self.name = None
        self.model = None
        self.orm_def = None

    def __repr__(self):
        return f'<Column {self.name}: {repr(self.type)}>'

    def __eq__(self, other):
        return self.orm_def == other

    def __ne__(self, other):
        return self.orm_def != other

    def __le__(self, other):
        return self.orm_def <= other

    def __lt__(self, other):
        return self.orm_def < other

    def __ge__(self, other):
        return self.orm_def >= other

    def __gt__(self, other):
        return self.orm_def > other


class Backref:
    def __init__(self, key: str, *, lazy: bool = True):
        self.key = key
        self.lazy = lazy

    def __repr__(self):
        return f'<Backref <- {self.key}' + (' lazy ' if self.lazy else '') + '>'


class Relationship:
    def __init__(self, model_name: str, *, backref: Backref = None, foreign_key: Column = None,
                 lazy: Union[bool, str] = False, secondary: str = None):
        self.model_name = model_name
        self.foreign_key = foreign_key
        self.backref = backref
        self.secondary = secondary
        self.lazy = lazy
        self.name = None
        self.model = None
        self.orm_def = None

    def __repr__(self):
        return f'<Relationship {self.name} -> {self.model_name}' + (' lazy ' if self.lazy else '') + '>'



from typing import Type

from bolinette import db


def model(model_name: str):
    def decorator(model_cls: Type['db.defs.Model']):
        db.models.register(model_name, model_cls)
        return model_cls

    return decorator


def mixin(mixin_name: str):
    def decorator(mixin_class: Type['db.defs.Mixin']):
        db.mixins.register(mixin_name, mixin_class)
        return mixin_class

    return decorator


def with_mixin(mixin_name: str):
    def decorator(model_cls):
        mixin_cls = db.mixins.get(mixin_name)
        for col_name, col_def in mixin_cls.columns().items():
            setattr(model_cls, col_name, col_def)
        for rel_name, rel_def in mixin_cls.relationships(model_cls).items():
            setattr(model_cls, rel_name, rel_def)
        return model_cls

    return decorator


def model_property(function):
    return db.defs.ModelProperty(function.__name__, function)



from bolinette.mapping.objects import Field, Column, List, Definition, Reference
from bolinette.mapping.definitions import register, get_payload, get_response
from bolinette.mapping.functions import map_model, validate_model, validate_payload, marshall, link_foreign_entities



from bolinette import mapping, env, db


@db.model('users_roles')
class UsersRoles(db.defs.Model):
    user_id = db.defs.Column(db.types.Integer, reference=db.defs.Reference('user', 'id'), primary_key=True)
    role_id = db.defs.Column(db.types.Integer, reference=db.defs.Reference('role', 'id'), primary_key=True)


@db.model('user')
class User(db.defs.Model):
    id = db.defs.Column(db.types.Integer, primary_key=True)
    username = db.defs.Column(db.types.String, unique=True, nullable=False)
    password = db.defs.Column(db.types.Password, nullable=False)
    email = db.defs.Column(db.types.Email, unique=env.init['USER_EMAIL_REQUIRED'],
                           nullable=(not env.init['USER_EMAIL_REQUIRED']))

    roles = db.defs.Relationship('role', secondary='users_roles', lazy='subquery',
                                 backref=db.defs.Backref('users', lazy=True))

    picture_id = db.defs.Column(db.types.Integer, reference=db.defs.Reference('file', 'id'))
    profile_picture = db.defs.Relationship('file', foreign_key=picture_id, lazy=False)

    timezone = db.defs.Column(db.types.String)

    @classmethod
    def payloads(cls):
        yield 'register', [
            mapping.Column(cls.username, required=True),
            mapping.Column(cls.email, required=True),
            mapping.Column(cls.password, required=True),
            mapping.Column(cls.timezone)
        ]
        yield 'admin_register', [
            mapping.Column(cls.username, required=True),
            mapping.Column(cls.email, required=True),
            mapping.Field(db.types.Boolean, name='send_mail', required=True)
        ]
        yield 'login', [
            mapping.Column(cls.username, required=True),
            mapping.Column(cls.password, required=True)
        ]

    @classmethod
    def responses(cls):
        default = [
            mapping.Column(cls.username),
            mapping.Reference(cls.profile_picture, 'minimal')
        ]
        yield default
        yield 'private', default + [
            mapping.Column(cls.email),
            mapping.List(mapping.Definition('role'), key='roles'),
            mapping.Column(cls.timezone)
        ]

    def __repr__(self):
        return f'<User {self.username}>'



from bolinette import ws

t = ws.Topic('user', login_required=True)


@t.subscribe
def user_topic_sub(topic, response, *, current_user, **_):
    topic.subscriptions[current_user.username] = [response]



from datetime import datetime
from typing import Type

from bolinette import db, mapping
from bolinette.exceptions import ParamConflictError, ParamMissingError, EntityNotFoundError


def marshall(definition, entity, *, skip_none=False, as_list=False, use_foreign_key=False):
    if not entity:
        return None
    if as_list:
        return [marshall(definition, e, skip_none=skip_none, as_list=False, use_foreign_key=use_foreign_key)
                for e in entity]
    data = {}
    for field in definition.fields:
        if isinstance(field, mapping.Field):
            if field.function is not None:
                value = field.function(entity)
            else:
                value = getattr(entity, field.key, None)
            if field.formatting is not None:
                value = field.formatting(value)
            if not skip_none or value is not None:
                data[field.name] = value
        elif isinstance(field, mapping.Reference) and use_foreign_key:
            data[field.foreign_key] = getattr(entity, field.foreign_key)
        elif isinstance(field, mapping.Definition):
            d = mapping.get_response(field.model_name, field.model_key)
            attr = None
            if field.function and callable(field.function):
                attr = field.function(entity)
            elif hasattr(entity, field.name):
                attr = getattr(entity, field.name)
            data[field.name] = marshall(d, attr, skip_none=skip_none, as_list=False, use_foreign_key=use_foreign_key)
        elif isinstance(field, mapping.List):
            d = mapping.get_response(field.element.model_name, field.element.model_key)
            data[field.name] = marshall(d, getattr(entity, field.name), skip_none=skip_none,
                                        as_list=True, use_foreign_key=use_foreign_key)
    return data


def link_foreign_entities(definition, params):
    errors = []
    for field in definition.fields:
        if isinstance(field, mapping.Reference):
            value = params.get(field.foreign_key, None)
            model = db.models.get(field.reference_model)
            if value is not None and model is not None:
                entity = model.query().filter_by(**{field.reference_key: value}).first()
                if entity is None:
                    errors.append((field.reference_model, field.reference_key, value))
    if len(errors) > 0:
        raise EntityNotFoundError(params=errors)


def validate_model(model: Type['db.defs.Model'], params: dict):
    errors = []
    for column in model.get_columns().values():
        key = column.name
        if column.primary_key:
            continue
        value = params.get(key, None)
        if column.unique and value is not None:
            if model.query().filter(column == value).first() is not None:
                errors.append((key, value))
    if len(errors) > 0:
        raise ParamConflictError(params=errors)
    return params


def validate_payload(definition, params, patch=False):
    errors = []
    valid = {}
    for field in definition.fields:
        if isinstance(field, mapping.Reference):
            if patch and field.foreign_key not in params:
                continue
            link = params.get(field.foreign_key)
            if not link or not len(str(link)):
                if field.required:
                    errors.append(field.foreign_key)
                else:
                    link = field.default
            valid[field.foreign_key] = link
        if isinstance(field, mapping.Field):
            if patch and field.name not in params:
                continue
            value = params.get(field.name, None)
            if value and field.type == db.types.Date:
                value = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
            if not value or not len(str(value)):
                if field.required:
                    errors.append(field.name)
                else:
                    value = field.default
            valid[field.name] = value
    if len(errors) > 0:
        raise ParamMissingError(params=errors)
    return valid


def map_model(model: Type['db.defs.Model'], entity, params, patch=False):
    errors = []
    for _, column in model.get_columns().items():
        key = column.name
        if column.primary_key or (key not in params and patch):
            continue
        original = getattr(entity, key)
        new = params.get(key, None)
        if original == new:
            continue
        if column.unique and new is not None:
            if model.query().filter(getattr(model, key) == new).first() is not None:
                errors.append((key, new))
        setattr(entity, key, new)
    if len(errors) > 0:
        raise ParamConflictError(params=errors)



from bolinette.db import types, defs
from bolinette.db.models import models
from bolinette.db.mixins import mixins
from bolinette.db.engine import engine
from bolinette.db.decorators import model, mixin, with_mixin, model_property



from bolinette.environment import env
from bolinette.bcrypt import bcrypt
import bolinette.models
import bolinette.controllers
import bolinette.topics
from bolinette.commands import command
from bolinette.bolinette import Bolinette



from typing import Type

from bolinette import mapping, db
from bolinette.exceptions import InternalError

_registered_responses = {}
_registered_payloads = {}


def _get_def(collection, model_name, key):
    m = collection.get(model_name)
    if m is None:
        raise InternalError(f'mapping.unknown_model:{model_name}')
    d = m.get(key)
    if d is None:
        raise InternalError(f'mapping.unknown_definition:{key}')
    return d


def get_response(model_name: str, key: str):
    return _get_def(_registered_responses, model_name, key)


def get_payload(model_name: str, key: str):
    return _get_def(_registered_payloads, model_name, key)


def register(model_name: str, model: Type['db.defs.Model']):
    def create_defs(collection, params):
        if params is None:
            return
        for param in params:
            if isinstance(param, list):
                model_key = 'default'
                payload = param
            else:
                model_key, payload = param
            definition = mapping.Definition(model_name, model_key)
            for field in payload:
                definition.fields.append(field)
            if definition.model_name not in collection:
                collection[definition.model_name] = {}
            collection[definition.model_name][definition.model_key] = definition

    create_defs(_registered_payloads, model.payloads())
    create_defs(_registered_responses, model.responses())



from typing import Type, Dict

from bolinette import db


class Mixins:
    def __init__(self):
        self.registered: Dict[str, Type['db.defs.Mixin']] = {}

    def get(self, mixin_name: str) -> Type['db.defs.Mixin']:
        return self.registered[mixin_name]

    def register(self, mixin_name: str, mixin_class: Type['db.defs.Mixin']):
        self.registered[mixin_name] = mixin_class


mixins = Mixins()



from aiohttp import web as aio_web
from aiohttp.web_request import Request

from bolinette import mapping
from bolinette.exceptions import ForbiddenError, APIError
from bolinette.network import transaction, AccessToken
from bolinette.web import deserialize, serialize, response
from bolinette.services import user_service
from bolinette.utils import Pagination


class Route:
    def __init__(self, *, func, base_url, path, method, access, expects, returns, roles):
        self.func = func
        self.base_url = base_url
        self.path = path
        self.method = method
        self.access = access
        self.expects = expects
        self.returns = returns
        self.roles = roles
        if self.roles and len(self.roles) and not self.access:
            self.access = AccessToken.Required

    async def handler(self, request: Request):
        current_user = None
        payload = await deserialize(request)
        match = {}
        query = {}
        for key in request.match_info:
            match[key] = request.match_info[key]
        for key in request.query:
            query[key] = request.query[key]
        try:
            with transaction:
                if self.access is not None:
                    current_user = await user_service.get_by_username(self.access.check(request))

                if self.roles and len(self.roles):
                    user_roles = set(map(lambda r: r.name, current_user.roles))
                    if 'root' not in user_roles and not len(user_roles.intersection(set(self.roles))):
                        raise ForbiddenError(f'user.forbidden:{",".join(self.roles)}')

                if self.expects is not None:
                    exp_def = mapping.get_payload(self.expects.model, self.expects.key)
                    payload = mapping.validate_payload(exp_def, payload, self.expects.patch)
                    mapping.link_foreign_entities(exp_def, payload)

                resp = await self.func(payload=payload, match=match, query=query, current_user=current_user)

            if isinstance(resp, aio_web.Response):
                return resp

            content = resp.content

            if content.get('data') is not None and isinstance(content['data'], Pagination):
                content['pagination'] = {
                    'page': content['data'].page,
                    'per_page': content['data'].per_page,
                    'total': content['data'].total,
                }
                content['data'] = content['data'].items

            if self.returns is not None:
                ret_def = mapping.get_response(self.returns.model, self.returns.key)
                if content.get('data') is not None:
                    content['data'] = mapping.marshall(ret_def, content['data'], skip_none=self.returns.skip_none,
                                                       as_list=self.returns.as_list)

            serialized, mime = serialize(content, 'application/json')

            web_response = aio_web.Response(text=serialized, status=resp.code, content_type=mime)
            for cookie in resp.cookies:
                if not cookie.delete:
                    web_response.set_cookie(cookie.name, cookie.value,
                                            expires=cookie.expires.strftime('%a, %d %b %Y %H:%M:%S GMT'),
                                            path=cookie.path, httponly=cookie.http_only)
                else:
                    web_response.del_cookie(cookie.name, path=cookie.path)

            return web_response
        except APIError as ex:
            res = response.from_exception(ex)
            serialized, mime = serialize(res.content, 'application/json')
            web_response = aio_web.Response(text=serialized, status=res.code, content_type=mime)
            return web_response

    def __str__(self):
        return f'<Route {self.method.http_verb} {self.base_url}{self.path} >'



from typing import Dict, Type

import sqlalchemy
from sqlalchemy import orm as sqlalchemy_orm

from bolinette import db


class Models:
    def __init__(self):
        self.registered: Dict[str, Type['db.defs.Model']] = {}

    def get(self, name: str) -> Type['db.defs.Model']:
        return self.registered.get(name, None)

    def register(self, model_name: str, model_cls: Type['db.defs.Model']):
        self.registered[model_name] = model_cls

    def init_models(self):
        orm_tables = {}
        for model_name, model_cls in self.registered.items():
            orm_cols = []
            for att_name, attribute in model_cls.get_columns().items():
                attribute.name = att_name
                attribute.model = model_cls
                ref = None
                if attribute.reference:
                    ref = sqlalchemy.ForeignKey(
                        f'{attribute.reference.model_name}.{attribute.reference.column_name}')
                attribute.orm_def = sqlalchemy.Column(att_name, attribute.type.sqlalchemy_type, ref,
                                                      primary_key=attribute.primary_key,
                                                      nullable=attribute.nullable,
                                                      unique=attribute.unique)
                orm_cols.append(attribute.orm_def)
            orm_tables[model_name] = sqlalchemy.Table(model_name, db.engine.model.metadata, *orm_cols)

        for model_name, model_cls in self.registered.items():
            orm_relationships = []
            for att_name, attribute in model_cls.get_relationships().items():
                attribute.name = att_name
                attribute.model = model_cls
                backref = None
                if attribute.backref:
                    backref = sqlalchemy_orm.backref(attribute.backref.key, lazy=attribute.backref.lazy)
                foreign_key = None
                if attribute.foreign_key:
                    foreign_key = attribute.foreign_key.orm_def
                secondary = None
                if attribute.secondary:
                    secondary = orm_tables[attribute.secondary]
                attribute.orm_def = sqlalchemy_orm.relationship(attribute.model_name, secondary=secondary,
                                                                lazy=attribute.lazy, foreign_keys=foreign_key,
                                                                backref=backref)
                orm_relationships.append(attribute)

            orm_defs = dict([(c.name, c.orm_def) for c in orm_relationships])
            orm_defs['__table__'] = orm_tables[model_name]
            orm_model = type(model_name, (db.engine.model,), orm_defs)

            for att_name, attribute in model_cls.get_properties().items():
                setattr(orm_model, att_name, property(attribute.function))

            setattr(self.registered[model_name], '__orm_model__', orm_model)
            setattr(self.registered[model_name], '__model_name__', model_name)


models = Models()
